package fe

import (
	"bufio"
	"bytes"
	"errors"
	"flag"
	"fmt"
	"html/template"
	"io"
	"log"
	"net"
	"net/http"
	"net/url"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"time"

	"git.autistici.org/ale/autoradio/Godeps/_workspace/src/github.com/gorilla/handlers"

	"git.autistici.org/ale/autoradio"
	"git.autistici.org/ale/autoradio/instrumentation"
)

var (
	proxyStreams = flag.Bool("enable-icecast-proxy", true, "Proxy the local icecast daemon")

	disableDebug  = flag.Bool("disable-debug", false, "Disable /debug/ URLs")
	restrictDebug = flag.Bool("restrict-debug", true, "Restrict access to /debug/ URLs to localhost")

	httpStatusCodes   = instrumentation.NewCounter("http.status")
	httpTargetStats   = instrumentation.NewCounter("http.target")
	sourceConnections = instrumentation.NewCounter("http.source_connections")
	source404Errors   = instrumentation.NewCounter("http.source_404")
	sourceErrors      = instrumentation.NewCounter("http.source_errors")
)

// ResponseWriter wrapper that logs an entry for every incoming HTTP
// request, as soon as the response headers are sent. This makes more
// sense for long-standing connections like in our case.
type logResponseWriter struct {
	http.ResponseWriter

	logged bool

	// Parameters of the original request.
	remoteAddr string
	method     string
	path       string
	proto      string
}

func (w *logResponseWriter) writeLog(code int) {
	log.Printf("%s %s %s %s %d", w.remoteAddr, w.method, w.path, w.proto, code)
	httpStatusCodes.IncrVar(strconv.Itoa(code))
	w.logged = true
}

func (w *logResponseWriter) WriteHeader(code int) {
	w.writeLog(code)
	w.ResponseWriter.WriteHeader(code)
}

func (w *logResponseWriter) Write(data []byte) (int, error) {
	// Catch the case when WriteHeader wasn't called.
	if !w.logged {
		w.writeLog(http.StatusOK)
	}
	return w.ResponseWriter.Write(data)
}

func (w *logResponseWriter) WrappedWriter() http.ResponseWriter {
	return w.ResponseWriter
}

func logHandler(h http.Handler) http.HandlerFunc {
	return func(w http.ResponseWriter, r *http.Request) {
		ip, _, err := net.SplitHostPort(r.RemoteAddr)
		if err != nil {
			ip = r.RemoteAddr
		}
		wrapw := &logResponseWriter{
			ResponseWriter: w,
			remoteAddr:     ip,
			method:         r.Method,
			path:           r.URL.Path,
			proto:          r.Proto,
		}
		h.ServeHTTP(wrapw, r)
	}
}

// HTTPRedirector makes clients talk to Icecast. This can happen
// either by sending HTTP redirects (if you want Icecast to be
// directly reachable from the outside), or by proxying the
// connections (and use just one open port).
type HTTPRedirector struct {
	domain    string
	staticDir string
	lb        *autoradioLoadBalancer
	client    *autoradio.Client
	template  *template.Template
	redirects map[string]string
}

// NewHTTPRedirector creates a new HTTP redirector.
func NewHTTPRedirector(client *autoradio.Client, domain, lbspec, staticDir, templateDir string) (*HTTPRedirector, error) {
	lb, err := parseLoadBalancerSpec(lbspec)
	if err != nil {
		return nil, err
	}
	tmpl := template.Must(
		template.ParseGlob(
			filepath.Join(templateDir, "*.html")))
	return &HTTPRedirector{
		client:    client,
		domain:    domain,
		lb:        lb,
		template:  tmpl,
		staticDir: staticDir,
	}, nil
}

// Pick a random IP with a protocol appropriate to the request (based
// on the remote address).
func randomIPForRequest(ips []net.IP, r *http.Request) net.IP {
	isV6 := false
	if host, _, err := net.SplitHostPort(r.RemoteAddr); err == nil {
		remoteAddr := net.ParseIP(host)
		isV6 = (remoteAddr != nil && (remoteAddr.To4() == nil))
	}
	return randomIPByProto(ips, isV6)
}

type httpRequestContext struct {
	req *http.Request
}

func (r *httpRequestContext) RemoteAddr() net.IP {
	host, _, _ := net.SplitHostPort(r.req.RemoteAddr)
	return net.ParseIP(host)
}

// Return an active node, chosen according to the current load
// balancing policy.
func (h *HTTPRedirector) pickActiveNode(r *http.Request) net.IP {
	result := h.lb.Choose(&httpRequestContext{r})
	if result == nil {
		return nil
	}
	return randomIPForRequest(result.IP, r)
}

// Periodically update the active nodes information in the load
// balancer. Requests will then use the cached data.
func (h *HTTPRedirector) lbUpdater() {
	for range time.NewTicker(2 * time.Second).C {
		nodes, err := h.client.GetNodes()
		if err != nil {
			continue
		}
		h.lb.Update(nodes)
	}
}

func icecastAddr(server net.IP) string {
	return net.JoinHostPort(server.String(), strconv.Itoa(autoradio.IcecastPort))
}

func streamURL(server net.IP, mountName string) string {
	var serverAddr string
	if *proxyStreams {
		serverAddr = server.String()
	} else {
		serverAddr = icecastAddr(server)
	}
	return fmt.Sprintf("http://%s%s", serverAddr, autoradio.MountNameToIcecastPath(mountName))
}

// Request wrapper that passes a Mount along with the HTTP request.
func (h *HTTPRedirector) withMount(f func(*autoradio.Mount, http.ResponseWriter, *http.Request)) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		mountPath := strings.TrimSuffix(r.URL.Path, ".m3u")
		mount, err := h.client.GetMount(mountPath)
		if err != nil {
			http.NotFound(w, r)
			return
		}
		f(mount, w, r)
	})
}

// Serve a M3U response. This simply points back at the stream
// redirect handler.
func (h *HTTPRedirector) serveM3U(mount *autoradio.Mount, w http.ResponseWriter, r *http.Request) {
	m3u := strings.TrimSuffix(r.URL.String(), ".m3u") + "\n"
	w.Header().Set("Content-Length", strconv.Itoa(len(m3u)))
	w.Header().Set("Content-Type", "audio/x-mpegurl")
	addDefaultHeaders(w)
	io.WriteString(w, m3u)
}

// Serve a response for a client connection to a relay.
func (h *HTTPRedirector) serveRelay(mount *autoradio.Mount, w http.ResponseWriter, r *http.Request) {
	// Find an active node.
	relayAddr := h.pickActiveNode(r)
	if relayAddr == nil {
		http.Error(w, "No active nodes", http.StatusServiceUnavailable)
		return
	}
	httpTargetStats.IncrVar(ipToMetric(relayAddr))

	// See if we need to serve a M3U response or a redirect.
	if strings.HasSuffix(r.URL.Path, ".m3u") {
		h.serveM3U(mount, w, r)
	} else {
		sendRedirect(w, r, streamURL(relayAddr, mount.Name))
	}
}

// Handle SOURCE requests.
func (h *HTTPRedirector) serveSource(mount *autoradio.Mount, w http.ResponseWriter, r *http.Request) {
	if mount.IsRelay() {
		log.Printf("source: connection to relay stream %s", mount.Name)
		http.Error(w, "Stream is relayed", http.StatusBadRequest)
		source404Errors.Incr()
		return
	}

	sourceConnections.Incr()

	// Find the current master node.
	masterInfo, err := h.client.GetMasterInfo()
	if err != nil {
		log.Printf("source error: no master: %v", err)
		http.Error(w, err.Error(), http.StatusServiceUnavailable)
		sourceErrors.Incr()
		return
	}

	// Create a ReverseProxy on the fly with the right backend
	// address. Blindly assume that the master has at least one
	// internal IP, and that we can reach it.
	target := icecastAddr(masterInfo.GetInternalIP()[0])
	tracker.setConnState(r, StateSource, target)
	proxy := &ReverseProxy{
		Director: func(req *http.Request) {
			req.URL.Scheme = "http"
			req.URL.Host = target
			req.URL.Path = autoradio.MountNameToIcecastPath(mount.Name)
		},
		FlushInterval: 500 * time.Millisecond,
	}
	proxy.ServeHTTP(w, r)
}

// Serve our cluster status page.
func (h *HTTPRedirector) serveStatusPage(w http.ResponseWriter, r *http.Request) {
	nodes, _ := h.client.GetNodes()
	mounts, _ := h.client.ListMounts()

	msl := autoradio.MountsToStatus(mounts, nodes)
	ctx := struct {
		Domain string
		Nodes  []*autoradio.NodeStatus
		Mounts []*autoradio.MountStatus
	}{h.domain, nodes, msl}

	var buf bytes.Buffer
	if err := h.template.ExecuteTemplate(&buf, "index.html", ctx); err != nil {
		log.Printf("error rendering template: %v", err)
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}
	w.Header().Set("Content-Type", "text/html; charset=utf-8")
	w.Header().Set("Content-Length", strconv.Itoa(buf.Len()))
	addDefaultHeaders(w)
	w.Write(buf.Bytes())
}

func withLocalhost(h http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		host, _, err := net.SplitHostPort(r.RemoteAddr)
		if err != nil {
			http.Error(w, "Unauthorized", http.StatusUnauthorized)
			return
		}
		if ip := net.ParseIP(host); !ip.IsLoopback() {
			http.Error(w, "Unauthorized", http.StatusUnauthorized)
			return
		}
		h.ServeHTTP(w, r)
	})
}

func (h *HTTPRedirector) createHandler() http.Handler {
	// Create our HTTP handler stack.
	mux := http.NewServeMux()

	// Serve static content.
	mux.Handle(
		"/static/",
		handlers.CompressHandler(
			http.StripPrefix(
				"/static/",
				http.FileServer(http.Dir(h.staticDir)))))

	if !*disableDebug {
		// Pass /debug/ to the default ServeMux, all the default debug
		// handlers are installed there. Add a debug handler for the
		// LoadBalancer data. Gzip the responses.
		debugMux := http.NewServeMux()
		debugMux.Handle("/debug/lbv2", h.lb)
		debugMux.Handle("/", http.DefaultServeMux)
		var h http.Handler = handlers.CompressHandler(debugMux)
		if *restrictDebug {
			h = withLocalhost(h)
		}
		mux.Handle("/debug/", h)
	}

	// Optionally enable a reverse proxy to the local Icecast for
	// the direct stream URLs (below IcecastMountPrefix).
	if *proxyStreams {
		iceProxy := NewSingleHostReverseProxy(&url.URL{Scheme: "http", Host: fmt.Sprintf("localhost:%d", autoradio.IcecastPort)})
		mux.HandleFunc(autoradio.IcecastMountPrefix+"/", func(w http.ResponseWriter, r *http.Request) {
			tracker.setConnState(r, StateListener, "localhost")
			iceProxy.ServeHTTP(w, r)
		})
	}

	// The root handler should discriminate between requests for
	// the status page, source and listener connections.
	relayHandler := h.withMount(h.serveRelay)
	sourceHandler := h.withMount(h.serveSource)
	statusPageHandler := handlers.CompressHandler(http.HandlerFunc(h.serveStatusPage))

	mux.HandleFunc("/", func(w http.ResponseWriter, r *http.Request) {
		switch {
		case r.Method == "SOURCE" || r.Method == "PUT":
			// Icecast 2.4 started supporting the PUT
			// method for sources, along the old SOURCE.
			sourceHandler.ServeHTTP(w, r)
		case r.URL.Path == "" || r.URL.Path == "/":
			statusPageHandler.ServeHTTP(w, r)
		default:
			relayHandler.ServeHTTP(w, r)
		}
	})

	// If a redirect map is present, run the redirect handler in
	// front of everything.
	var rooth http.Handler = mux
	if h.redirects != nil {
		rooth = h.redirectHandler(rooth)
	}

	// Instrument the resulting HTTP handler, and map global
	// redirects.
	return logHandler(trackRequestsHandler(rooth))
}

// Create a handler that will detect requests matching our redirect
// map, and will serve a redirect for them.
func (h *HTTPRedirector) redirectHandler(wrap http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		if target, ok := h.redirects[r.URL.Path]; ok {
			sendRedirect(w, r, target)
			return
		}
		wrap.ServeHTTP(w, r)
	})
}

func parseRedirects(r io.Reader) (map[string]string, error) {
	out := make(map[string]string)
	s := bufio.NewScanner(r)
	for s.Scan() {
		line := s.Text()
		if line == "" || strings.HasPrefix(line, "#") {
			continue
		}
		fields := strings.Fields(line)
		if len(fields) != 2 {
			return nil, errors.New("syntax error")
		}
		out[fields[0]] = fields[1]
	}
	return out, nil
}

func (h *HTTPRedirector) LoadStaticRedirects(filename string) error {
	f, err := os.Open(filename)
	if err != nil {
		return err
	}
	defer f.Close()
	h.redirects, err = parseRedirects(f)
	return err
}

// Run starts the HTTP server on the given addr. Does not return.
func (h *HTTPRedirector) Run(addr string) {
	// Start the background goroutine that updates the
	// LoadBalancer asynchronously.
	go h.lbUpdater()

	httpServer := &http.Server{
		Addr:         addr,
		Handler:      h.createHandler(),
		ReadTimeout:  10 * time.Second,
		WriteTimeout: 10 * time.Second,
	}

	log.Printf("starting HTTP server on %s/tcp", httpServer.Addr)
	log.Fatal(httpServer.ListenAndServe())
}

func addDefaultHeaders(w http.ResponseWriter) {
	w.Header().Set("Expires", "-1")
	w.Header().Set("Cache-Control", "no-store")
}

func sendRedirect(w http.ResponseWriter, r *http.Request, target string) {
	// Firefox apparently caches redirects regardless of
	// the status code, so we have to add some quite
	// aggressive cache-busting headers. We serve a status
	// code of 307 to HTTP/1.1 clients, 302 otherwise.
	w.Header().Set("Cache-Control", "max-age=0,no-cache,no-store")
	w.Header().Set("Pragma", "no-cache")
	w.Header().Set("Expires", "-1")
	code := http.StatusFound
	if r.ProtoMinor == 1 {
		code = http.StatusTemporaryRedirect
	}
	http.Redirect(w, r, target, code)
}
